import numpy
import six

if six.PY3:
    from typing import Sequence

from mpilot.libraries.eems.exceptions import MixedArrayShapes


class SameArrayShapeMixin(object):
    def validate_array_shapes(self, arrays, lineno=None):
        # type: (Sequence[numpy.ma.masked_array], int) -> None

        if len(arrays) == 1:
            return

        shape = arrays[0].shape

        for arr in arrays:
            if arr.shape != shape:
                raise MixedArrayShapes(shape, arr.shape, lineno)



from collections import namedtuple

from ply import yacc, lex
from ply.lex import TOKEN

ProgramNode = namedtuple("ProgramNode", ("commands", "version"))
CommandNode = namedtuple("CommandNode", ("result_name", "command", "arguments", "lineno"))
ArgumentNode = namedtuple("ArgumentNode", ("name", "value", "lineno"))
ExpressionNode = namedtuple("ExpressionNode", ("value", "lineno"))


class Lexer(object):
    def __init__(self):
        self.lexer = lex.lex(module=self)

    tokens = [
        "COLON",
        "COMMA",
        "EQUAL",
        "FALSE",
        "FLOAT",
        "ID",
        "INT",
        "LBRACK",
        "LPAREN",
        "PLAIN_STRING",
        "RBRACK",
        "RPAREN",
        "STRING",
        "TRUE",
    ]

    t_ignore = " \t"
    t_ignore_COMMENT = r'\#.*'

    t_COLON = ":"
    t_COMMA = ","
    t_EQUAL = "="
    t_FALSE = "False"
    t_LBRACK = r"\["
    t_LPAREN = r"\("
    t_PLAIN_STRING = r"[^\#\:\,\=\(\)\[\]\"\r\n]+"
    t_RBRACK = r"\]"
    t_RPAREN = r"\)"
    t_TRUE = "True"

    @TOKEN(r"[a-zA-Z_][a-zA-Z_0-9]*")
    def t_ID(self, t):
        return t

    @TOKEN(r"[\-\+]?((\d+\.\d*)|(\.\d+))([eE][\+\-]?\d+)?")
    def t_FLOAT(self, t):
        t.value = float(t.value)
        return t

    @TOKEN(r"[\-\+]?\d+")
    def t_INT(self, t):
        t.value = int(t.value)
        return t

    @TOKEN(r'("(\\.|[^"\\])*")|(\'(\\.|[^"\\])*\')')
    def t_STRING(self, t):
        t.value = t.value.strip("\"'").encode().decode("unicode_escape")
        return t

    @TOKEN(r"[\r\n]+")
    def t_newline(self, t):
        t.lexer.lineno += len(t.value)

    def t_error(self, t):
        raise SyntaxError("Illegal character {0} at position {1}".format(t.value[0], t.lexpos))


class Parser(object):
    tokens = Lexer.tokens

    def __init__(self):
        self.parser = yacc.yacc(module=self, debug=False)
        self.lexer = Lexer().lexer
        self.eems_v2 = False

    def p_program(self, p):
        """
        program : commands
        """

        p[0] = ProgramNode(p[1], 2 if self.eems_v2 else 3)

    def p_commands(self, p):
        """
        commands : command commands
        """

        p[0] = [p[1]] + p[2]

    def p_commands_command(self, p):
        """
        commands : command
        """

        p[0] = [p[1]]

    def p_command(self, p):
        """
        command : ID EQUAL ID arguments
        """

        p[0] = CommandNode(p[1], p[3], p[4], p.lineno(3))

    def p_eems2_command(self, p):
        """
        command : ID arguments
        """

        self.eems_v2 = True
        p[0] = CommandNode(None, p[1], p[2], p.lineno(1))

    def p_arguments(self, p):
        """
        arguments : LPAREN argument_list RPAREN
        """

        p[0] = p[2]

    def p_argument_empty(self, p):
        """
        arguments : LPAREN RPAREN
        """

        p[0] = []

    def p_argument_list(self, p):
        """
        argument_list : argument COMMA argument_list
        """

        p[0] = [p[1]] + p[3]

    def p_argument_list_argument(self, p):
        """
        argument_list : argument COMMA
                      | argument
        """
        p[0] = [p[1]]

    def p_argument(self, p):
        """
        argument : ID EQUAL expression
        """

        p[0] = ArgumentNode(p[1], p[3], p.lineno(1))

    def p_expression(self, p):
        """
        expression : ID
                   | plain_string
                   | STRING
                   | number
                   | list
                   | boolean
        """

        p[0] = ExpressionNode(p[1], p.lineno(1))

    def p_expression_identifier_expression(self, p):
        """
        expression : ID expression
        """

        p[0] = ExpressionNode(p[1] + str(p[2].value), p.lineno(1))

    def p_plain_string(self, p):
        """
        plain_string : PLAIN_STRING
                     | ID
        """

        p[0] = p[1]

    def p_plain_string_with_colon(self, p):
        """
        plain_string : plain_string COLON plain_string
        """

        p[0] = p[1] + ":" + p[3]

    def p_number(self, p):
        """
        number : INT
               | FLOAT
        """

        p[0] = p[1]

    def p_list(self, p):
        """
        list : LBRACK elements RBRACK
        """

        p[0] = p[2]

    def p_list_empty(self, p):
        """
        list : LBRACK RBRACK
        """

        p[0] = []

    def p_elements(self, p):
        """
        elements : element COMMA elements
        """

        p[0] = [p[1]] + p[3]

    def p_elements_element(self, p):
        """
        elements : element COMMA
                 | element
        """

        p[0] = [p[1]]

    def p_element_expression(self, p):
        """
        element : expression
        """

        p[0] = p[1]

    def p_element_tuple_pairs(self, p):
        """
        elements : tuple_pairs
        """

        p[0] = p[1]

    def p_tuple_pairs(self, p):
        """
        tuple_pairs : tuple_pair COMMA tuple_pairs
        """

        p[0] = dict(list(p[3].items()) + [p[1]])

    def p_tuple_pairs_pair(self, p):
        """
        tuple_pairs : tuple_pair COMMA
                    | tuple_pair
        """

        p[0] = dict([p[1]])

    def p_tuple_pair(self, p):
        """
        tuple_pair : STRING COLON expression
                   | PLAIN_STRING COLON expression
                   | ID COLON expression
        """

        p[0] = (p[1], p[3])

    def p_boolean(self, p):
        """
        boolean : TRUE
                | FALSE
        """

        p[0] = p[1]

    def p_error(self, p):
        if p:
            raise SyntaxError("Syntax error '{0}' at position {1}".format(p.value, p.lexpos))
        else:
            raise SyntaxError("Invalid syntax at end of statement")

    def parse(self, source):
        # type: (str) -> ProgramNode
        """ Parses the source text into a program structure """

        return self.parser.parse(source, lexer=self.lexer, tracking=True)



import six
from six import python_2_unicode_compatible

if six.PY3:
    from typing import Union, Sequence, Set, Any


class MPilotError(Exception):
    pass


@python_2_unicode_compatible
class ProgramError(MPilotError):
    def __init__(self, lineno=None, message=None):
        # type: (int, str) -> None

        self.lineno = lineno
        self.message = message

    def __str__(self):
        return self.message or "ProgramError"


@python_2_unicode_compatible
class CommandDoesNotExist(ProgramError):
    """ The command file contains a command that could not be found in the command libraries. """

    def __init__(self, name, lineno=None):
        # type: (str, int) -> None

        super(CommandDoesNotExist, self).__init__(lineno)

        self.name = name

    def __str__(self):
        return "\n".join(
            (
                'Problem: The command "{}" does not exist.'.format(self.name),
                "Solution: Make sure the command exists and is spelled correctly.",
            )
        )


@python_2_unicode_compatible
class DuplicateResult(ProgramError):
    """ The same result name was used more than once. """

    def __init__(self, result, lineno=None):
        # type: (str, int) -> None

        super(DuplicateResult, self).__init__(lineno)

        self.result = result

    def __str__(self):
        return "\n".join(
            (
                'Problem: The result name "{}" is duplicated in the command file.'.format(
                    self.result
                ),
                "Solution: Make sure that the result name for each command doesn't exist elsewhere in the model.",
            )
        )


@python_2_unicode_compatible
class MissingParameters(ProgramError):
    """ A required command argument is missing. """

    def __init__(self, command, parameters, lineno=None):
        # type: (Union[str, Any], Union[Sequence[str], Set[str]], int) -> None

        super(MissingParameters, self).__init__(lineno)

        self.command = (
            command if isinstance(command, six.string_types) else command.name
        )
        self.parameters = parameters

    def __str__(self):
        return "\n".join(
            (
                'Problem: The command "{}" is missing the following required parameters: {}'.format(
                    self.command, ", ".join(self.parameters)
                ),
                "Solution: Make sure all required parameters are provided for this command.",
            )
        )


@python_2_unicode_compatible
class NoSuchParameter(ProgramError):
    """ An argument was provided that is not valid for the command. """

    def __init__(self, command, parameter, lineno=None):
        # type: (Union[str, Any], str, int) -> None

        super(NoSuchParameter, self).__init__(lineno)

        self.command = (
            command if isinstance(command, six.string_types) else command.name
        )
        self.parameter = parameter

    def __str__(self):
        return "\n".join(
            (
                'Problem: The command "{}" has no parameter named "{}".'.format(
                    self.command, self.parameter
                ),
                "Solution: Make sure the parameters are correct for this command.",
            )
        )


@python_2_unicode_compatible
class ParameterNotValid(ProgramError):
    """ The value is invalid for the given argument. """

    def __init__(self, value, required_type, lineno=None):
        # type: (Any, str, int) -> None

        super(ParameterNotValid, self).__init__(lineno)

        self.value = value
        self.required_type = required_type

    def __str__(self):
        return "\n".join(
            (
                "Problem: A value of type {} was expected, but value {} of type {} was provided.".format(
                    self.required_type,
                    six.text_type(self.value),
                    six.text_type(type(self.value)),
                ),
                "Solution: Make sure command parameters are correct.",
            )
        )


@python_2_unicode_compatible
class PathDoesNotExist(ProgramError):
    """ The path is required to exist, but doesn't. """

    def __init__(self, path, lineno=None):
        # type: (str, int) -> None

        super(PathDoesNotExist, self).__init__(lineno)

        self.path = path

    def __str__(self):
        return "\n".join(
            (
                "Problem: The following path does not exist: {}".format(self.path),
                "Solution: Double check the path and create it if necessary.",
            )
        )


@python_2_unicode_compatible
class InvalidRelativePath(ProgramError):
    """ A relative path was provided, but the program doesn't have a working directory """

    def __init__(self, path, lineno=None):
        # type: (str, int) -> None

        super(InvalidRelativePath, self).__init__(lineno)

        self.path = path

    def __str__(self):
        return "\n".join(
            (
                "Problem: The following path is relative, but no working directory is set: {}".format(
                    self.path
                ),
                "Solution: Use absolute paths for this command file.",
            )
        )


@python_2_unicode_compatible
class ResultDoesNotExist(ProgramError):
    """ A result was referenced which does not exist. """

    def __init__(self, result, lineno=None):
        # type: (str, int) -> None

        super(ResultDoesNotExist, self).__init__(lineno)

        self.result = result

    def __str__(self):
        return "\n".join(
            (
                'Problem: The result "{}" does not exist.'.format(self.result),
                "Solution: Double check that the result has been defined elsewhere in the command file.",
            )
        )


@python_2_unicode_compatible
class ResultTypeNotValid(ProgramError):
    """ A specified result is not valid for the argument. """

    def __init__(self, result, lineno=None):
        # type: (str, int) -> None

        super(ResultTypeNotValid, self).__init__(lineno)

        self.result = result

    def __str__(self):
        return "\n".join(
            (
                'Problem: The value returned by "{}" does not match the required parameter type for this parameter.'.format(
                    self.result
                ),
                "Solution: Double check that the correct command result is being used for this parameter.",
            )
        )


@python_2_unicode_compatible
class RecursiveModelStructure(ProgramError):
    """ The model graph is contains circular references. """

    def __str__(self):
        return "\n".join(
            (
                "Problem: The model is recursive.",
                "Solution: Double check that command references don't create a loop",
            )
        )



from __future__ import absolute_import

import os
import sys

import click
import six

from ..exceptions import MPilotError, ProgramError
from ..program import Program, EEMS_CSV_LIBRARIES, EEMS_NETCDF_LIBRARIES

LINE_CONTEX_LENGTH = 3


@click.command()
@click.argument("library")
@click.argument("path")
@click.option(
    "--library",
    "-l",
    "libraries",
    multiple=True,
    default=[],
    help="Add a command library by its Python path (e.g., mpilot.libraries.eems.csv)",
)
def main(library, path, libraries):
    if not os.path.exists(path):
        sys.stderr.write(
            "\n".join(
                (
                    "Problem: The specified command file does not exist: {}".format(
                        path
                    ),
                    "Solution: Check the path to the command file and try again.",
                )
            )
        )
        sys.stderr.write("\n")
        sys.exit(-1)

    with open(path) as f:
        lines = [line.strip("\n\r") for line in f.readlines()]
    source = "\n".join(lines)

    if not libraries:
        libraries = tuple()

    try:
        program = Program.from_source(
            source,
            libraries=libraries
            + (EEMS_CSV_LIBRARIES if library == "eems-csv" else EEMS_NETCDF_LIBRARIES),
            working_dir=os.path.dirname(path),
        )
        program.run()
    except MPilotError as ex:
        sys.stderr.write(
            "\n".join(
                (
                    "ERROR: There was a problem running the MPilot command file.",
                    six.text_type(ex),
                )
            )
        )
        sys.stderr.write("\n")

        if isinstance(ex, ProgramError) and ex.lineno is not None:
            idx = ex.lineno - 1
            start = max(idx - LINE_CONTEX_LENGTH, 0)
            end = min(idx + LINE_CONTEX_LENGTH, len(lines))

            sys.stderr.write("\n".join((" " * 4 + line for line in lines[start:idx])))
            sys.stderr.write("\n")
            sys.stderr.write("--> {}".format(lines[idx]))
            sys.stderr.write("\n")
            sys.stderr.write(
                "\n".join((" " * 4 + line for line in lines[idx + 1 : end]))
            )
            sys.stderr.write("\n")

        sys.exit(-1)




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'COLON COMMA EQUAL FALSE FLOAT ID INT LBRACK LPAREN PLAIN_STRING RBRACK RPAREN STRING TRUE\n        program : commands\n        \n        commands : command commands\n        \n        commands : command\n        \n        command : ID EQUAL ID arguments\n        \n        command : ID arguments\n        \n        arguments : LPAREN argument_list RPAREN\n        \n        arguments : LPAREN RPAREN\n        \n        argument_list : argument COMMA argument_list\n        \n        argument_list : argument COMMA\n                      | argument\n        \n        argument : ID EQUAL expression\n        \n        expression : ID\n                   | plain_string\n                   | STRING\n                   | number\n                   | list\n                   | boolean\n        \n        expression : ID expression\n        \n        plain_string : PLAIN_STRING\n                     | ID\n        \n        plain_string : plain_string COLON plain_string\n        \n        number : INT\n               | FLOAT\n        \n        list : LBRACK elements RBRACK\n        \n        list : LBRACK RBRACK\n        \n        elements : element COMMA elements\n        \n        elements : element COMMA\n                 | element\n        \n        element : expression\n        \n        elements : tuple_pairs\n        \n        tuple_pairs : tuple_pair COMMA tuple_pairs\n        \n        tuple_pairs : tuple_pair COMMA\n                    | tuple_pair\n        \n        tuple_pair : STRING COLON expression\n                   | PLAIN_STRING COLON expression\n                   | ID COLON expression\n        \n        boolean : TRUE\n                | FALSE\n        '
    
_lr_action_items = {'ID':([0,3,6,7,8,11,14,15,16,17,19,29,33,40,46,47,48,49,50,],[4,4,9,-5,13,-7,-4,-6,13,19,19,40,44,19,40,55,19,19,19,]),'$end':([1,2,3,5,7,11,14,15,],[0,-1,-3,-2,-5,-7,-4,-6,]),'EQUAL':([4,13,],[6,17,]),'LPAREN':([4,9,],[8,8,]),'RPAREN':([8,10,12,16,18,19,20,21,22,23,24,25,26,27,28,30,31,32,35,43,44,45,],[11,15,-10,-9,-8,-12,-11,-13,-14,-15,-16,-17,-19,-22,-23,-37,-38,-18,-25,-21,-20,-24,]),'COMMA':([12,19,20,21,22,23,24,25,26,27,28,30,31,32,35,36,38,39,40,41,42,43,44,45,56,57,58,],[16,-12,-11,-13,-14,-15,-16,-17,-19,-22,-23,-37,-38,-18,-25,46,-29,47,-12,-14,-19,-21,-20,-24,-36,-34,-35,]),'STRING':([17,19,29,40,46,47,48,49,50,],[22,22,41,22,41,53,22,22,22,]),'PLAIN_STRING':([17,19,29,33,40,46,47,48,49,50,],[26,26,42,26,26,42,54,26,26,26,]),'INT':([17,19,29,40,46,48,49,50,],[27,27,27,27,27,27,27,27,]),'FLOAT':([17,19,29,40,46,48,49,50,],[28,28,28,28,28,28,28,28,]),'LBRACK':([17,19,29,40,46,48,49,50,],[29,29,29,29,29,29,29,29,]),'TRUE':([17,19,29,40,46,48,49,50,],[30,30,30,30,30,30,30,30,]),'FALSE':([17,19,29,40,46,48,49,50,],[31,31,31,31,31,31,31,31,]),'RBRACK':([19,21,22,23,24,25,26,27,28,29,30,31,32,34,35,36,37,38,39,40,41,42,43,44,45,46,47,51,52,56,57,58,],[-12,-13,-14,-15,-16,-17,-19,-22,-23,35,-37,-38,-18,45,-25,-28,-30,-29,-33,-12,-14,-19,-21,-20,-24,-27,-32,-26,-31,-36,-34,-35,]),'COLON':([19,21,26,40,41,42,43,44,53,54,55,],[-20,33,-19,48,49,50,33,-20,49,50,48,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'program':([0,],[1,]),'commands':([0,3,],[2,5,]),'command':([0,3,],[3,3,]),'arguments':([4,9,],[7,14,]),'argument_list':([8,16,],[10,18,]),'argument':([8,16,],[12,12,]),'expression':([17,19,29,40,46,48,49,50,],[20,32,38,32,38,56,57,58,]),'plain_string':([17,19,29,33,40,46,48,49,50,],[21,21,21,43,21,21,21,21,21,]),'number':([17,19,29,40,46,48,49,50,],[23,23,23,23,23,23,23,23,]),'list':([17,19,29,40,46,48,49,50,],[24,24,24,24,24,24,24,24,]),'boolean':([17,19,29,40,46,48,49,50,],[25,25,25,25,25,25,25,25,]),'elements':([29,46,],[34,51,]),'element':([29,46,],[36,36,]),'tuple_pairs':([29,46,47,],[37,37,52,]),'tuple_pair':([29,46,47,],[39,39,39,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> program","S'",1,None,None,None),
  ('program -> commands','program',1,'p_program','parser.py',84),
  ('commands -> command commands','commands',2,'p_commands','parser.py',91),
  ('commands -> command','commands',1,'p_commands_command','parser.py',98),
  ('command -> ID EQUAL ID arguments','command',4,'p_command','parser.py',105),
  ('command -> ID arguments','command',2,'p_eems2_command','parser.py',112),
  ('arguments -> LPAREN argument_list RPAREN','arguments',3,'p_arguments','parser.py',120),
  ('arguments -> LPAREN RPAREN','arguments',2,'p_argument_empty','parser.py',127),
  ('argument_list -> argument COMMA argument_list','argument_list',3,'p_argument_list','parser.py',134),
  ('argument_list -> argument COMMA','argument_list',2,'p_argument_list_argument','parser.py',141),
  ('argument_list -> argument','argument_list',1,'p_argument_list_argument','parser.py',142),
  ('argument -> ID EQUAL expression','argument',3,'p_argument','parser.py',148),
  ('expression -> ID','expression',1,'p_expression','parser.py',155),
  ('expression -> plain_string','expression',1,'p_expression','parser.py',156),
  ('expression -> STRING','expression',1,'p_expression','parser.py',157),
  ('expression -> number','expression',1,'p_expression','parser.py',158),
  ('expression -> list','expression',1,'p_expression','parser.py',159),
  ('expression -> boolean','expression',1,'p_expression','parser.py',160),
  ('expression -> ID expression','expression',2,'p_expression_identifier_expression','parser.py',167),
  ('plain_string -> PLAIN_STRING','plain_string',1,'p_plain_string','parser.py',174),
  ('plain_string -> ID','plain_string',1,'p_plain_string','parser.py',175),
  ('plain_string -> plain_string COLON plain_string','plain_string',3,'p_plain_string_with_colon','parser.py',182),
  ('number -> INT','number',1,'p_number','parser.py',189),
  ('number -> FLOAT','number',1,'p_number','parser.py',190),
  ('list -> LBRACK elements RBRACK','list',3,'p_list','parser.py',197),
  ('list -> LBRACK RBRACK','list',2,'p_list_empty','parser.py',204),
  ('elements -> element COMMA elements','elements',3,'p_elements','parser.py',211),
  ('elements -> element COMMA','elements',2,'p_elements_element','parser.py',218),
  ('elements -> element','elements',1,'p_elements_element','parser.py',219),
  ('element -> expression','element',1,'p_element_expression','parser.py',226),
  ('elements -> tuple_pairs','elements',1,'p_element_tuple_pairs','parser.py',233),
  ('tuple_pairs -> tuple_pair COMMA tuple_pairs','tuple_pairs',3,'p_tuple_pairs','parser.py',240),
  ('tuple_pairs -> tuple_pair COMMA','tuple_pairs',2,'p_tuple_pairs_pair','parser.py',247),
  ('tuple_pairs -> tuple_pair','tuple_pairs',1,'p_tuple_pairs_pair','parser.py',248),
  ('tuple_pair -> STRING COLON expression','tuple_pair',3,'p_tuple_pair','parser.py',255),
  ('tuple_pair -> PLAIN_STRING COLON expression','tuple_pair',3,'p_tuple_pair','parser.py',256),
  ('tuple_pair -> ID COLON expression','tuple_pair',3,'p_tuple_pair','parser.py',257),
  ('boolean -> TRUE','boolean',1,'p_boolean','parser.py',264),
  ('boolean -> FALSE','boolean',1,'p_boolean','parser.py',265),
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mpilot',
 'mpilot.cli',
 'mpilot.libraries',
 'mpilot.libraries.eems',
 'mpilot.libraries.eems.csv',
 'mpilot.libraries.eems.netcdf',
 'mpilot.parser']

package_data = \
{'': ['*']}

install_requires = \
['click<8.0', 'packaging', 'ply>=3.11,<4.0', 'six']

extras_require = \
{':python_version >= "2.7" and python_version < "2.8"': ['numpy<=1.16.6'],
 ':python_version >= "3.6"': ['numpy>=1.19.4,<2.0.0'],
 'netcdf': ['netCDF4<1.5.6']}

entry_points = \
{'console_scripts': ['mpilot = mpilot.cli.mpilot:main']}

setup_kwargs = {
    'name': 'mpilot',
    'version': '1.0.0b16',
    'description': 'MPilot is a plugin-based, environmental modeling framework',
    'long_description': '# MPilot (beta)\nMPilot is a plugin-based, environmental modeling framework based on a bottom-up, many-to-many workflow that can be \nrepresented by a directed (not iterating) graph. MPilot is descended from the Environmental Evaluation Modeling System \n(EEMS), which was initially a fuzzy logic modeling package based on EMDS.\n\n[MPilot Documentation](https://consbio.github.io/mpilot/)\n\n# Installing\n\nMPilot with EEMS can be installed with `pip`:\n\n```bash\n$ pip install mpilot\n```\n\nIn order to run MPilot with NetCDF datasets, you\'ll need to install the NetCDF variant:\n\n```bash\n$ pip install mpilot[netcdf]\n```\n\n# Creating models\nMPilot models are contained in "command files", using a simple scripting language. Here is an example model, which \nloads two columns of integer data from a CSV file, sums them, and writes the result to a second CSV file.\n\n```text\nA = EEMSRead(\n    InFileName = "input.csv",\n    InFieldName = "A",\n    DataType = "Integer"\n)\nB = EEMSRead(\n    InFileName = "input.csv",\n    InFieldName = "B",\n    DataType = "Integer"\n)\nAPlusB = Sum(\n    InFieldNames = [A, B]\n)\nOut = EEMSWrite(\n    OutFileName = "output.csv",\n    OutFieldNames = [A, B, APlusB]\n)\n```\n\n# Running models\n\nModels are run using the included `mpilot` program. The following commands will run a model using the EEMS CSV library \nand using the EEMS NetCDF library respectively:\n\n```bash\n$ mpilot eems-csv model.mpt\n$ mpilot eems-netcdf model.mpt\n```\n',
    'author': 'Conservation Biology Institute',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://consbio.github.io/mpilot/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)



from __future__ import absolute_import

import numpy
from netCDF4 import Dataset
from numpy.ma import is_masked

from mpilot import params
from mpilot.commands import Command
from mpilot.utils import insure_fuzzy
from .exceptions import NoSuchVariable, InvalidPositiveData, InvalidFuzzyData

FUZZY_MIN = -1
FUZZY_MAX = 1


class EEMSRead(Command):
    """ Reads a variable from a file, converting floats to nearest int when necessary. """

    display_name = "Read"
    inputs = {
        "InFileName": params.PathParameter(must_exist=True),
        "InFieldName": params.StringParameter(),
        "MissingValue": params.NumberParameter(required=False),
        "DataType": params.DataTypeParameter(
            required=False,
            valid_types={
                "Float": numpy.float64,
                "Integer": numpy.int,
                "Positive Float": numpy.float64,
                "Positive Integer": numpy.uint,
                "Fuzzy": numpy.float64,
            },
        ),
    }
    output = params.DataParameter()

    def execute(self, **kwargs):
        path = kwargs["InFileName"]
        variable_name = kwargs["InFieldName"]
        data_type = kwargs.get("DataType", numpy.float64)

        with Dataset(kwargs["InFileName"], "r") as dataset:
            if kwargs["InFieldName"] not in dataset.variables:
                raise NoSuchVariable(path, variable_name, lineno=self.lineno)

        variable = dataset[variable_name]
        data = variable[:]

        if (
            self.arguments.get("DataType", "Float")
            in ("Positive Integer", "Positive Float")
            and data.min() < 0
        ):
            raise InvalidPositiveData(
                path, self.arguments["DataType"], lineno=self.lineno
            )

        if numpy.issubdtype(data.dtype, numpy.float) and data_type in (
            numpy.int,
            numpy.uint,
        ):
            data = numpy.rint(data, out=data)  # round in-place

        result = numpy.ma.array(
            data,
            mask=data.mask if is_masked(data) else False,
            dtype=data_type,
            fill_value=999999 if data_type in (numpy.int, numpy.uint) else None,
        )

        if self.arguments.get("DataType", "Float") == "Fuzzy":
            fuzzy_pad = 0.01 * (FUZZY_MAX - FUZZY_MIN)

            if data.max() > FUZZY_MAX + fuzzy_pad or data.min() < FUZZY_MIN - fuzzy_pad:
                raise InvalidFuzzyData(path, lineno=self.lineno)

            insure_fuzzy(result)

        result.soften_mask()

        # Mask missing values
        if "MissingValue" in kwargs:
            missing_value = (
                int(kwargs["MissingValue"])
                if numpy.issubdtype(result.mask.dtype, numpy.int)
                else float(kwargs["MissingValue"])
            )

            self.result.mask = numpy.where(
                result.data == missing_value, True, result.mask or False
            )

        result.data[result.mask] = result.fill_value

        return result


class EEMSWrite(Command):
    """ Writes one or more file """

    display_name = "Write"
    inputs = {
        "OutFileName": params.PathParameter(),
        "OutFieldNames": params.ListParameter(
            params.ResultParameter(params.DataParameter())
        ),
        "DimensionFileName": params.PathParameter(must_exist=True),
        "DimensionFieldName": params.StringParameter(),
    }
    output = params.BooleanParameter()

    def execute(self, **kwargs):
        commands = kwargs["OutFieldNames"]
        arrays = [c.result for c in commands]
        self.validate_array_shapes(arrays)

        with Dataset(kwargs["OutFileName"], "w") as dataset:
            with Dataset(kwargs["DimensionsFileName"]) as dim_dataset:
                dimensions = dim_dataset[kwargs["DimensionFieldName"]].dimensions
                for dimension in dimensions:
                    in_variable = dim_dataset[dimension]
                    out_variable = dataset.createDimension(dimension, in_variable.size)

                    for attribute in dir(in_variable):
                        if attribute not in dir(out_variable) and attribute not in (
                            "_FillValue",
                            "missing_value",
                        ):
                            setattr(
                                out_variable, attribute, getattr(in_variable, attribute)
                            )
                    out_variable[:] = in_variable[:]

            mask = numpy.copy(arrays[0].mask)
            for arr in arrays[1:]:
                mask |= arr.mask

            for command in commands:
                variable = dataset.createVariable(
                    command.result_name,
                    command.result.dtype.char,
                    dimensions,
                    fill_value=command.result.fill_value,
                )
                variable[:] = numpy.ma.MaskedArray(command.result.data, mask)

        return True



from __future__ import print_function

from collections import namedtuple

import six
from six import add_metaclass

if six.PY3:
    from typing import List, Any, Dict

from mpilot.exceptions import MissingParameters, NoSuchParameter
from mpilot.params import TupleParameter


Argument = namedtuple("Argument", ("name", "value", "lineno"))
CommandInfo = namedtuple("CommandInfo", ("module", "command"))


class CommandMeta(type):
    """ Command metaclass, used to automatically register commands, which can then be loaded by name. """

    _commands = set()

    def __new__(mcs, name, bases, attrs):
        attrs.update(
            {"inputs": attrs.get("inputs", {}), "output": attrs.get("output", None)}
        )

        # Add an optional "Metadata" input to all commands
        attrs["inputs"]["Metadata"] = TupleParameter(required=False)

        new_class = super(CommandMeta, mcs).__new__(mcs, name, bases, attrs)

        # Use an explicit command name if one is set, otherwise use the class name
        command_name = attrs.get("name", name)

        new_class.name = command_name
        new_class.display_name = attrs.get("display_name", command_name)
        new_class.required_inputs = {
            name: param for name, param in attrs["inputs"].items() if param.required
        }
        new_class.allow_extra_inputs = attrs.get("allow_extra_inputs", False)

        if not any(
            info.module == new_class.__module__
            and getattr(info.command, "name", info.command.__name__) == command_name
            for info in mcs._commands
        ):
            mcs._commands.add(CommandInfo(new_class.__module__, new_class))

        new_class._commands = mcs._commands

        return new_class


@add_metaclass(CommandMeta)
class Command(object):
    @classmethod
    def get_commands(cls):
        # type: () -> List[CommandInfo]

        return list(cls._commands)

    def __init__(self, result_name, arguments=[], program=None, lineno=None):
        # type: (str, List[Any], Any, int) -> None

        self.result_name = result_name
        self.arguments = arguments
        self.program = program
        self.lineno = lineno

        self.argument_lines = {arg.name: arg.lineno for arg in arguments}

        self.is_finished = False
        self._result = None

    @property
    def result(self):
        if not self.is_finished:
            self.run()

        return self._result

    @property
    def metadata(self):
        # type: () -> Dict[str, str]

        for arg in self.arguments:
            if arg.name == "Metadata":
                return self.inputs["Metadata"].clean(
                    arg.value, self.program, self.argument_lines.get(arg.name)
                )
        return {}

    def get_argument_value(self, name, default=None):
        # type: (str, Any) -> Any

        for arg in self.arguments:
            if arg.name == name:
                return arg.value
        return default

    def validate_params(self, params):
        # type: (Dict[str, Any]) -> Dict[str, Any]

        required_inputs = [key for key, value in self.inputs.items() if value.required]
        all_inputs = [key for key in self.inputs.keys()]

        missing_params = set(required_inputs).difference(params.keys())
        if missing_params:
            raise MissingParameters(self.result_name, missing_params, self.lineno)

        cleaned = {}
        for name, value in params.items():
            if name not in all_inputs:
                if not self.allow_extra_inputs:
                    raise NoSuchParameter(
                        self.result_name, name, self.argument_lines.get(name)
                    )
                else:
                    cleaned[name] = value

            else:
                cleaned[name] = self.inputs[name].clean(
                    value, self.program, self.argument_lines.get(name)
                )

        return cleaned

    def run(self):
        if not self.is_finished:
            self.is_running = True
            self._result = self.execute(
                **self.validate_params({arg.name: arg.value for arg in self.arguments})
            )
            self.is_finished = True

    def execute(self, **kwargs):
        raise NotImplemented
